--
-- JCB320T  -- Hardcoded --
--
-- @author:    	Xentro (Marcus@Xentro.se)
-- @website:	www.Xentro.se
-- @history:	v1.0 - 2015-01-05 - 
-- 

JCB320T = {};

JCB320T.LEFT = "LEFT";
JCB320T.RIGHT = "RIGHT";

-- Animations --
JCB320T.START_UP = "startUp";
JCB320T.SHUT_DOWN = "shutDown";
JCB320T.FUEL_NEEDLE = "fuelDisplay";
JCB320T.WIPER_BUTTON = "wiperButton";

function JCB320T.initSpecialization(specializations)
    Vehicle.registerJointType("skidSteer");
end;

function JCB320T.prerequisitesPresent(specializations)
    return true;
end;

function JCB320T:load(xmlFile)
	self.updateDirtyObject = SpecializationUtil.callSpecializationsFunction("updateDirtyObject");
	self.quickHitchAnimation = SpecializationUtil.callSpecializationsFunction("quickHitchAnimation");
	self.getPtoRpm = Utils.overwrittenFunction(self.getPtoRpm, JCB320T.getPtoRpm);
	self.getDoConsumePtoPower = Utils.overwrittenFunction(self.getDoConsumePtoPower, JCB320T.getDoConsumePtoPower);
	
	self.igntionStuff = {};
	self.igntionStuff.startUpSequence = false;
	self.igntionStuff.fuelNeedleSequence = false;
	self.igntionStuff.onEnter = false;
	self.igntionStuff.lastMode = 0;
	self.igntionStuff.lastFuel = 0;
	
	self.joysticks = {};
	self.joysticks["LIMIT"] = Utils.degToRad(getXMLFloat(xmlFile, "vehicle.SkidSteer.joysticks#rotLimit"));
	self.joysticks["RIGHT"] = {};
	self.joysticks["RIGHT"].node = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.SkidSteer.joysticks#right"));
	self.joysticks["RIGHT"].startRot = {getRotation(self.joysticks["RIGHT"].node)};
	self.joysticks["RIGHT"].lastRot = {0, 0};
	
	self.joysticks["LEFT"] = {};
	self.joysticks["LEFT"].node = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.SkidSteer.joysticks#left"));
	self.joysticks["LEFT"].startRot = {getRotation(self.joysticks["LEFT"].node)};
	self.joysticks["LEFT"].lastRot = {0, 0};
	
	self.rightStickDirtyFlag = self:getNextDirtyFlag();
	self.leftStickDirtyFlag = self:getNextDirtyFlag();
	
	local wiper = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.SkidSteer.wiper#index"));
	if wiper ~= nil then
		self.wiper = {};
		self.wiper.node = wiper;
		self.wiper.active = false;
		self.wiper.finished = true;
		
		local animCharSet = 0;
		animCharSet = getAnimCharacterSet(wiper);
		if animCharSet ~= 0 then
			self.wiper.animCharSet = animCharSet;
			local clip = getAnimClipIndex(animCharSet, getXMLString(xmlFile, "vehicle.SkidSteer.wiper#clip"));
			
			if clip ~= nil then
				assignAnimTrackClip(animCharSet, 0, clip);
				setAnimTrackLoopState(animCharSet, 0, true);
				self.wiper.speed = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.SkidSteer.wiper#speedScale"), 1);
				setAnimTrackSpeedScale(animCharSet, 0, self.wiper.speed);
				self.wiper.duration = getAnimClipDuration(animCharSet, clip);
			end;
		end;
	end;
	
	local seatNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.SkidSteer.seat#index"));
	if seatNode ~= nil then
		self.seat = {};
		self.seat.node = seatNode;
		self.seat.wheelId = getXMLInt(xmlFile, "vehicle.SkidSteer.seat#wheelReferenceId") or 1;
		self.seat.usagePercentage = (getXMLInt(xmlFile, "vehicle.SkidSteer.seat#usagePercentage") or 50) / 100;
		self.seat.offsetY = getXMLFloat(xmlFile, "vehicle.SkidSteer.seat#offsetY");
	end;
	
	local quickNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.SkidSteer.quickHitch#index"));
	if quickNode ~= nil then
		self.quickHitch = {};
		self.quickHitch.node = quickNode;
		self.quickHitch.activate = false;
		self.quickHitch.dir = 1;
		self.quickHitch.lastDir = 1;
		
		local animCharSet = getAnimCharacterSet(quickNode);
		if animCharSet ~= 0 then
			self.quickHitch.animCharSet = animCharSet;
			local clip = getAnimClipIndex(animCharSet, getXMLString(xmlFile, "vehicle.SkidSteer.quickHitch#clip"));
			
			if clip ~= nil then
				assignAnimTrackClip(animCharSet, 0, clip);
				setAnimTrackLoopState(animCharSet, 0, false);
				self.quickHitch.duration = getAnimClipDuration(animCharSet, clip);
			end;
		end;
	end;
	
	self.workRpm = {};
	self.workRpm.rpm = 0;
	self.workRpm.releaseRpmTimer = 0;
	self.workRpm.maxTime = 300;
	self.workRpm.rpmOnLoad = 400;
	self.workRpm.movingToolIds = {3, 4};
	
	self.forceRoadLightMirrorChange = false;
	
	self.movingPartsNameFix = {1, 3, 8, 9};
end;

function JCB320T:delete()
end;

function JCB320T:readStream(streamId, connection)
end;

function JCB320T:writeStream(streamId, connection)
end;

function JCB320T:readUpdateStream(streamId, timestamp, connection)
	if streamReadBool(streamId) then
		self.joysticks["LEFT"].lastRot[1] = streamReadFloat32(streamId);
		self.joysticks["LEFT"].lastRot[2] = streamReadFloat32(streamId);
		self:updateDirtyObject(JCB320T.LEFT, self.joysticks["LEFT"].lastRot[1], self.joysticks["LEFT"].lastRot[2]);
		
		if not connection:getIsServer() then
			self:raiseDirtyFlags(self.leftStickDirtyFlag);
		end;
	end;
	
	if streamReadBool(streamId) then
		self.joysticks["RIGHT"].lastRot[1] = streamReadFloat32(streamId);
		self.joysticks["RIGHT"].lastRot[2] = streamReadFloat32(streamId);
		self:updateDirtyObject(JCB320T.RIGHT, self.joysticks["RIGHT"].lastRot[1], self.joysticks["RIGHT"].lastRot[2]);
		
		if not connection:getIsServer() then
			self:raiseDirtyFlags(self.rightStickDirtyFlag); -- update clients
		end;
	end;
end;
   
function JCB320T:writeUpdateStream(streamId, connection, dirtyMask)
	local s = (connection:getIsServer() or connection ~= self.owner); -- either we are on the client, or the target connection is not the owner
	
	if bitAND(dirtyMask, self.leftStickDirtyFlag) ~= 0 and s then
		streamWriteBool(streamId, true);
		streamWriteFloat32(streamId, self.joysticks["LEFT"].lastRot[1]);
		streamWriteFloat32(streamId, self.joysticks["LEFT"].lastRot[2]);
	else
		streamWriteBool(streamId, false);
	end;
	
	if bitAND(dirtyMask, self.rightStickDirtyFlag) ~= 0 and s then
		streamWriteBool(streamId, true);
		streamWriteFloat32(streamId, self.joysticks["RIGHT"].lastRot[1]);
		streamWriteFloat32(streamId, self.joysticks["RIGHT"].lastRot[2]);
	else
		streamWriteBool(streamId, false);
	end;
end;

function JCB320T:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
	if not resetVehicles then
		local rot1 = getXMLFloat(xmlFile, key .. "#rot1");
		local rot2 = getXMLFloat(xmlFile, key .. "#rot2");
		local rot3 = getXMLFloat(xmlFile, key .. "#rot3");
		local rot4 = getXMLFloat(xmlFile, key .. "#rot4");
		
		if rot1 ~= nil then
			setRotation(self.movingParts[self.movingPartsNameFix[1]].node, rot1, 0, 0);
			setRotation(self.movingParts[self.movingPartsNameFix[2]].node, rot2, 0, 0);
			setRotation(self.movingParts[self.movingPartsNameFix[3]].node, rot3, 0, 0);
			setRotation(self.movingParts[self.movingPartsNameFix[4]].node, rot4, 0, 0);
			
			self:setMovingToolDirty(self.movingTools[3].node);
		end;
	end;
	
	return BaseMission.VEHICLE_LOAD_OK;
end;

function JCB320T:getSaveAttributesAndNodes(nodeIdent)
	local rot1 = {getRotation(self.movingParts[self.movingPartsNameFix[1]].node)};
	local rot2 = {getRotation(self.movingParts[self.movingPartsNameFix[2]].node)};
	local rot3 = {getRotation(self.movingParts[self.movingPartsNameFix[3]].node)};
	local rot4 = {getRotation(self.movingParts[self.movingPartsNameFix[4]].node)};
	local attributes = 'rot1="' .. rot1[1] ..'" rot2="' .. rot2[1] ..'" rot3="' .. rot3[1] ..'" rot4="' .. rot4[1] .. '"';
	
	return attributes;
end;

function JCB320T:mouseEvent(posX, posY, isDown, isUp, button)
end;

function JCB320T:keyEvent(unicode, sym, modifier, isDown)
end;

function JCB320T:update(dt)
	if not self.firstJCBRun then
		self.firstJCBRun = true;
		self:setMovingToolDirty(self.movingTools[5].node); -- update hose on load 
	end;
	
	if self:getIsActive() then		
		if self:getIsActiveForInput(false) and self.isClient then
			local AS = -self.axisSide;
			local AF = self.axisForward;
			
			-- only allow right/left if forward/backward is active
			if AF == 0 and AS ~= 0 then
				AS = 0;
			end;
			
			local newSideRot = -self.joysticks["LEFT"].startRot[3] + (self.joysticks["LIMIT"] * AS);
			local newForwardRot = -self.joysticks["LEFT"].startRot[1] + (self.joysticks["LIMIT"] * AF);
			
			if (self.joysticks["LEFT"].lastRot[1] ~= -newForwardRot or self.joysticks["LEFT"].lastRot[2] ~= newSideRot) then
				self.joysticks["LEFT"].lastRot[1] = -newForwardRot;
				self.joysticks["LEFT"].lastRot[2] = newSideRot;
				
				self:updateDirtyObject(JCB320T.LEFT, self.joysticks["LEFT"].lastRot[1], self.joysticks["LEFT"].lastRot[2]);
				self:raiseDirtyFlags(self.leftStickDirtyFlag);
			end;
		
			local arm, armType = InputBinding.getInputAxis(self.movingTools[3].axisActionIndex);
			local tool, toolType = InputBinding.getInputAxis(self.movingTools[4].axisActionIndex);
			
			arm = -arm;
			
			if armType == InputBinding.INPUTTYPE_MOUSE_AXIS then
				arm = math.max(math.min(arm, 1), -1);
			end;
			if toolType == InputBinding.INPUTTYPE_MOUSE_AXIS then
				tool = math.max(math.min(tool, 1), -1);
			end;
			
			local newArmRot = self.joysticks["RIGHT"].startRot[1] + (self.joysticks["LIMIT"] * arm);
			local newToolRot = self.joysticks["RIGHT"].startRot[3] + (self.joysticks["LIMIT"] * tool);
		
			if (self.joysticks["RIGHT"].lastRot[1] ~= newArmRot or self.joysticks["RIGHT"].lastRot[2] ~= newToolRot) then
				self.joysticks["RIGHT"].lastRot[1] = newArmRot;
				self.joysticks["RIGHT"].lastRot[2] = newToolRot;
				
				self:updateDirtyObject(JCB320T.RIGHT, self.joysticks["RIGHT"].lastRot[1], self.joysticks["RIGHT"].lastRot[2]);
				self:raiseDirtyFlags(self.rightStickDirtyFlag); -- update server
			end;
		end;
		
		if self.isServer then
			-- keep tool attached synched!
			if self.movingTools[3].lastRotSpeed ~= 0 then
				local implementIndex = self:getImplementIndexByJointDescIndex(1);
				
				if implementIndex ~= nil then
					self:setMovingToolDirty(self.movingTools[1].node);
				end;
			end;
		end;
		
		if self.seat ~= nil and self.wheels[self.seat.wheelId] ~= nil then
			if self.seat.trans == nil then
				self.seat.trans = {getTranslation(self.seat.node)};
				
				if self.seat.offsetY ~= nil then
					self.seat.trans[2] = self.seat.trans[2] + self.seat.offsetY;
				end;
			end;
			
			local _, y, _ = getTranslation(self.wheels[self.seat.wheelId].repr);
			local newTrans = (y * self.seat.usagePercentage) + self.seat.trans[2];
			newTrans = math.floor(newTrans * 10^3 + 0.5) / 10^3;
			
			if newTrans ~= self.seat.lastTrans then
				self.ikChainsById["rightFoot"].isDirty = true;
				self.ikChainsById["leftFoot"].isDirty = true;
				
				setTranslation(self.seat.node, self.seat.trans[1], newTrans, self.seat.trans[3]);
				self.seat.lastTrans = newTrans;
			end;
		end;
		
		if self.quickHitch ~= nil and self.quickHitch.activate then
			self:setMovingToolDirty(self.movingTools[5].node);
		end;
	end;
end;

function JCB320T:updateTick(dt)
	if self:getIsActive() then
		local ignitionMod = self.ignitionMode ~= nil;
		
		if ignitionMod then
			if self.key.mode ~= self.igntionStuff.lastMode then
				if self.igntionStuff.lastMode == 0 then
					self:playAnimation(JCB320T.START_UP, 1);
					self.igntionStuff.startUpSequence = true;
				elseif self.key.mode == 0 then
					self:playAnimation(JCB320T.SHUT_DOWN, 1);
					self:playAnimation(JCB320T.FUEL_NEEDLE, -1);
				end;
				
				self.igntionStuff.lastMode = self.key.mode;
			end;
		end;
		
		if self.igntionStuff.startUpSequence then
			if self:getAnimationTime(JCB320T.START_UP) >= 1 then
				self.igntionStuff.startUpSequence = false;
				
				self:playAnimation(JCB320T.SHUT_DOWN, 1);
				self:setAnimationStopTime(JCB320T.SHUT_DOWN, 0.25);
				
				self:playAnimation(JCB320T.FUEL_NEEDLE, -1);
				self:setAnimationStopTime(JCB320T.FUEL_NEEDLE, self.fuelFillLevel / self.fuelCapacity);
				self.igntionStuff.fuelNeedleSequence = true;
			end;
		else		
			if (ignitionMod and self.key.mode ~= 0 or not ignitionMod) then
				if self.igntionStuff.fuelNeedleSequence then
					if self:getAnimationTime(JCB320T.FUEL_NEEDLE) <= self.fuelFillLevel / self.fuelCapacity then
						self.igntionStuff.fuelNeedleSequence = false;
						
						if self:getIsAnimationPlaying(JCB320T.FUEL_NEEDLE) then
							self:stopAnimation(JCB320T.FUEL_NEEDLE);
						end;
					end;
				else
					local fuelF = math.floor(self.fuelFillLevel);
					
					if self.igntionStuff.lastFuel ~= fuelF then
						self:setAnimationTime(JCB320T.FUEL_NEEDLE, self.fuelFillLevel / self.fuelCapacity, true);
						self.igntionStuff.lastFuel = fuelF;
					end;
				end;
			end;
		end;
		
		local neededRpm = 0;
		local implementIndex = self:getImplementIndexByJointDescIndex(1);
		
		if implementIndex ~= nil then
			local obj = self.attachedImplements[implementIndex].object;
			
			if obj.getPtoRpm ~= nil then
				local rpm = obj:getPtoRpm()
				
				if rpm > 0 then
					neededRpm = self.workRpm.rpmOnLoad + rpm;
				end;
			end;
		end;
		
		if neededRpm == 0 then
			for _, toolId in ipairs(self.workRpm.movingToolIds) do
				local tool = self.movingTools[toolId];
				
				if tool.rotSpeed ~= nil then
					if tool.lastRotHistoryRPM == nil then
						tool.lastRotHistoryRPM = tool.curRot[tool.rotationAxis];
					end;
					
					if 0.00001 < math.abs(tool.lastRotHistoryRPM - tool.curRot[tool.rotationAxis]) then
						tool.lastRotHistoryRPM = tool.curRot[tool.rotationAxis];
						neededRpm = self.workRpm.rpmOnLoad;
					end;
				end;
				
				if tool.transSpeed ~= nil then
					if tool.lastTransHistoryRPM == nil then
						tool.lastTransHistoryRPM = tool.curTrans[tool.translationAxis];
					end;
					
					if 0.00001 < math.abs(tool.lastTransHistoryRPM - tool.curTrans[tool.translationAxis]) then
						tool.lastTransHistoryRPM = tool.curTrans[tool.translationAxis];
						neededRpm = self.workRpm.rpmOnLoad;
					end;
				end;
			end;
		end;
		
		if neededRpm > 0 then
			self.workRpm.rpm = neededRpm;
			self.workRpm.releaseRpmTimer = g_currentMission.time + self.workRpm.maxTime;
		else
			if g_currentMission.time > self.workRpm.releaseRpmTimer then
				self.workRpm.rpm = 0;
				self.workRpm.releaseRpmTimer = 0;
			end;
		end;
	end;
	
	if self.wiper ~= nil and self.wiper.animCharSet ~= nil then
		if not self.wiper.finished then
			if getAnimTrackTime(self.wiper.animCharSet, 0) % self.wiper.duration <= 100 then
				setAnimTrackTime(self.wiper.animCharSet, 0, 0.0);
				disableAnimTrack(self.wiper.animCharSet, 0);			
				self.wiper.finished = true;
			end;			
		end;
		
		local allow = (self.ignitionMode == nil and self.isMotorStarted or self.ignitionMode ~= nil and self.key.mode ~= 0);
		local g = g_currentMission.environment;
		if g.currentRain ~= nil and g.currentRain.rainTypeId == "rain" and g.lastRainScale > 0.01 and allow then
			if not self.wiper.active then
				enableAnimTrack(self.wiper.animCharSet, 0);
				self:playAnimation(JCB320T.WIPER_BUTTON, 1);
				self.wiper.active = true;
			end;
		else
			if self.wiper.active then
				self.wiper.active = false;
				self:playAnimation(JCB320T.WIPER_BUTTON, -1);
				self.wiper.finished = false;
			end;
		end;
	end;
	
	if self.quickHitch ~= nil and self.quickHitch.activate then
		local currentTime = getAnimTrackTime(self.quickHitch.animCharSet, 0);
		local stopClip = false;
		
		if currentTime > self.quickHitch.duration then
			setAnimTrackTime(self.quickHitch.animCharSet, 0, self.quickHitch.duration);
			stopClip = true;
		elseif currentTime < 0 then
			setAnimTrackTime(self.quickHitch.animCharSet, 0, 0);
			stopClip = true;
		end;
		
		if stopClip then
			disableAnimTrack(self.quickHitch.animCharSet, 0);
			self.quickHitch.activate = false;
		end;
	end;
end;

function JCB320T:draw()
end;

function JCB320T:onEnter()
	if self.ignitionMode == nil then
		self:playAnimation(JCB320T.START_UP, 1);
		self.igntionStuff.startUpSequence = true;
	end;
end;

function JCB320T:onLeave()
	if self.ignitionMode == nil then
		self:playAnimation(JCB320T.SHUT_DOWN, 1);
		self:playAnimation(JCB320T.FUEL_NEEDLE, -1);
	end;
	
	self.workRpm.rpm = 0;
	self.workRpm.releaseRpmTimer = 0;
	
	self.forceRoadLightMirrorChange = true;
end;

function JCB320T:attachImplement()
	self:quickHitchAnimation(1);
end;

function JCB320T:detachImplement()
	self:quickHitchAnimation(-1);
end;

function JCB320T:updateDirtyObject(t, x, z)
	if t == JCB320T.RIGHT then
		self.ikChainsById["rightArm"].isDirty = true;
		self.ikChainsById["rubberRight"].isDirty = true;
	else -- LEFT
		self.ikChainsById["leftArm"].isDirty = true;
		self.ikChainsById["rubberLeft"].isDirty = true;
	end;
	
	setRotation(self.joysticks[t].node, x, 0, z);
end;

function JCB320T:quickHitchAnimation(dir)
	if self.quickHitch ~= nil then
		setAnimTrackSpeedScale(self.quickHitch.animCharSet, 0, dir);
		enableAnimTrack(self.quickHitch.animCharSet, 0);
		self.quickHitch.activate = true;
	end;
end;

function JCB320T:getPtoRpm(oldFunc)
	return self.workRpm.rpm;
	
	-- local rpm = self.workRpm.rpm;
	
	-- if oldFunc ~= nil then
		-- rpm = rpm + oldFunc(self);
	-- end;
	
	-- return rpm;
end;

function JCB320T:getDoConsumePtoPower(oldFunc)
	return self.workRpm.rpm > 0;
	
	-- if not consume and oldFunc ~= nil then
		-- return oldFunc(self);
	-- end;
	
	-- return consume;
end;



if not Utils.JCBSkidSteerFix then
	Utils.JCBSkidSteerFix = true;
	local name = "skidSteers"
	
	if StoreItemsUtil.storeCategories[name] == nil then
		local numCat = 0;
		
		for catName in pairs(StoreItemsUtil.storeCategories) do
			numCat = numCat + 1;
		end;
		
		StoreItemsUtil.storeCategories[name] = {orderId = numCat + 1, name = name, title = g_i18n:getText(name), image = g_currentModDirectory .. "skidsteers/store_JCB.png"};
	end;
end;